"""
GitHub Events Monitor package

Provides:
- FastAPI REST API (see `api.py`)
- MCP server (see `mcp_server.py`)
- Core collector and data access (see `collector.py`)
"""

__all__ = [
	"api",
	"mcp_server",
	"collector",
	"config",
]



